import textwrap

from strawberry.schema_codegen import codegen


def test_codegen_object_type():
    schema = """
    type Example {
        a: Int!
        b: Float!
        c: Boolean!
        d: String!
        e: ID!
        f: [Int!]!
        g: [Float!]!
        h: [Boolean!]!
        i: [String!]!
        j: [ID!]!
        k: [Int]
        l: [Float]
        m: [Boolean]
        n: [String]
        o: [ID]
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Example:
            a: int
            b: float
            c: bool
            d: str
            e: strawberry.ID
            f: list[int]
            g: list[float]
            h: list[bool]
            i: list[str]
            j: list[strawberry.ID]
            k: list[int | None] | None
            l: list[float | None] | None
            m: list[bool | None] | None
            n: list[str | None] | None
            o: list[strawberry.ID | None] | None
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_supports_descriptions():
    schema = """
    "Example description"
    type Example {
        "a description"
        a: Int!
        "b description"
        b: Float!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type(description="Example description")
        class Example:
            a: int = strawberry.field(description="a description")
            b: float = strawberry.field(description="b description")
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_supports_interfaces():
    schema = """
    interface Node {
        id: ID!
    }

    type User implements Node {
        id: ID!
        name: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.interface
        class Node:
            id: strawberry.ID

        @strawberry.type
        class User(Node):
            id: strawberry.ID
            name: str
        """
    ).strip()

    assert codegen(schema).strip() == expected



import textwrap

from strawberry.schema_codegen import codegen


def test_adds_schema_if_schema_is_defined():
    schema = """
    type Root {
        hello: String!
    }

    type RootMutation {
        hello: String!
    }

    type RootSubscription {
        hello: String!
    }

    schema {
        query: Root
        mutation: RootMutation
        subscription: RootSubscription
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Root:
            hello: str

        @strawberry.type
        class RootMutation:
            hello: str

        @strawberry.type
        class RootSubscription:
            hello: str

        schema = strawberry.Schema(query=Root, mutation=RootMutation, subscription=RootSubscription)
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_adds_schema_if_has_query():
    schema = """
    type Query {
        hello: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Query:
            hello: str

        schema = strawberry.Schema(query=Query)
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_adds_schema_if_has_mutation():
    schema = """
    type Mutation {
        hello: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Mutation:
            hello: str

        schema = strawberry.Schema(mutation=Mutation)
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_adds_schema_if_has_subscription():
    schema = """
    type Subscription {
        hello: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Subscription:
            hello: str

        schema = strawberry.Schema(subscription=Subscription)
        """
    ).strip()

    assert codegen(schema).strip() == expected



import textwrap
from dataclasses import dataclass
from textwrap import dedent
from typing import Annotated, Any, ClassVar, Generic, TypeVar, Union

import pytest

import strawberry
from strawberry.exceptions import InvalidUnionTypeError
from strawberry.types.lazy_type import lazy


def test_union_as_field():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class Query:
        ab: A | B = strawberry.field(default_factory=lambda: A(a=5))

    schema = strawberry.Schema(query=Query)
    query = """{
        ab {
            __typename,

            ... on A {
                a
            }
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["ab"] == {"__typename": "A", "a": 5}


def test_union_as_field_inverse():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class Query:
        ab: A | B = strawberry.field(default_factory=lambda: B(b=5))

    schema = strawberry.Schema(query=Query)
    query = """{
        ab {
            __typename,

            ... on B {
                b
            }
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["ab"] == {"__typename": "B", "b": 5}


def test_cannot_use_non_strawberry_fields_for_the_union():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class Query:
        ab: A | B = "ciao"

    schema = strawberry.Schema(query=Query)
    query = """{
        ab {
            __typename,

            ... on A {
                a
            }
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert (
        result.errors[0].message
        == 'The type "<class \'str\'>" cannot be resolved for the field "ab" '
        ", are you using a strawberry.field?"
    )


def test_union_as_mutation_return():
    @strawberry.type
    class A:
        x: int

    @strawberry.type
    class B:
        y: int

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def hello(self) -> A | B:
            return B(y=5)

    schema = strawberry.Schema(query=A, mutation=Mutation)

    query = """
    mutation {
        hello {
            __typename

            ... on A {
                x
            }

            ... on B {
                y
            }
        }
    }
    """

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["hello"] == {"__typename": "B", "y": 5}


def test_types_not_included_in_the_union_are_rejected():
    @strawberry.type
    class Outside:
        c: int

    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def hello(self) -> A | B:
            return Outside(c=5)  # type:ignore

    schema = strawberry.Schema(query=A, mutation=Mutation, types=[Outside])

    query = """
    mutation {
        hello {
            __typename

            ... on A {
                a
            }

            ... on B {
                b
            }
        }
    }
    """

    result = schema.execute_sync(query)

    assert (
        result.errors[0].message == "The type "
        "\"<class 'tests.schema.test_union.test_types_not_included_in_the_union_are_rejected.<locals>.Outside'>\""
        ' of the field "hello" '
        "is not in the list of the types of the union: \"['A', 'B']\""
    )


def test_unknown_types_are_rejected():
    @strawberry.type
    class Outside:
        c: int

    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> A | B:
            return Outside(c=5)  # type:ignore

    schema = strawberry.Schema(query=Query)

    query = """
    {
        hello {
            ... on A {
                a
            }
        }
    }
    """

    result = schema.execute_sync(query)

    assert "Outside" in result.errors[0].message


def test_named_union():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    Result = Annotated[A | B, strawberry.union(name="Result")]

    @strawberry.type
    class Query:
        ab: Result = strawberry.field(default_factory=lambda: A(a=5))

    schema = strawberry.Schema(query=Query)

    query = """{
        __type(name: "Result") {
            kind
            description
        }

        ab {
            __typename,

            ... on A {
                a
            }
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["ab"] == {"__typename": "A", "a": 5}
    assert result.data["__type"] == {"kind": "UNION", "description": None}


def test_named_union_description():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    Result = Annotated[
        A | B, strawberry.union(name="Result", description="Example Result")
    ]

    @strawberry.type
    class Query:
        ab: Result = strawberry.field(default_factory=lambda: A(a=5))

    schema = strawberry.Schema(query=Query)

    query = """{
        __type(name: "Result") {
            kind
            description
        }

        ab {
            __typename,

            ... on A {
                a
            }
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["ab"] == {"__typename": "A", "a": 5}
    assert result.data["__type"] == {"kind": "UNION", "description": "Example Result"}


def test_can_use_union_in_optional():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    Result = Annotated[A | B, strawberry.union(name="Result")]

    @strawberry.type
    class Query:
        ab: Result | None = None

    schema = strawberry.Schema(query=Query)

    query = """{
        __type(name: "Result") {
            kind
            description
        }

        ab {
            __typename,

            ... on A {
                a
            }
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["ab"] is None


def test_optional_annotated_union_preserves_metadata():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    Result = Annotated[A | B | None, strawberry.union(name="Result")]

    @strawberry.type
    class Query:
        ab: Result = None

    schema = strawberry.Schema(query=Query)

    query = """
    {
        __type(name: "Result") {
            kind
            possibleTypes {
                name
            }
        }

        __schema {
            queryType {
                fields {
                    name
                    type {
                        kind
                        name
                        ofType {
                            name
                        }
                    }
                }
            }
        }
    }
    """

    result = schema.execute_sync(query)

    assert not result.errors

    # Check the union exists with the correct name
    union_type = result.data["__type"]
    assert union_type is not None, (
        "Expected introspection to return a type for 'Result'"
    )
    assert union_type["kind"] == "UNION"

    # Check union members are correct (A and B)
    possible_types = {t["name"] for t in union_type["possibleTypes"]}
    assert possible_types == {"A", "B"}

    # Check the field "ab" uses the correct union type
    fields = result.data["__schema"]["queryType"]["fields"]
    ab_field = next(f for f in fields if f["name"] == "ab")

    # Because it's Optional, GraphQL type is nullable -> no NON_NULL wrapper
    field_type = ab_field["type"]

    # The field is Optional, so it must NOT be NON_NULL
    assert field_type["kind"] != "NON_NULL"

    assert field_type["name"] == "Result"


def test_multiple_unions():
    @strawberry.type
    class CoolType:
        @strawberry.type
        class UnionA1:
            value: int

        @strawberry.type
        class UnionA2:
            value: int

        @strawberry.type
        class UnionB1:
            value: int

        @strawberry.type
        class UnionB2:
            value: int

        field1: UnionA1 | UnionA2
        field2: UnionB1 | UnionB2

    schema = strawberry.Schema(query=CoolType)

    query = """
        {
            __type(name:"CoolType") {
                name
                description
                fields {
                    name
                }
            }
        }
    """

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["__type"] == {
        "description": None,
        "fields": [{"name": "field1"}, {"name": "field2"}],
        "name": "CoolType",
    }


def test_union_used_multiple_times():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    MyUnion = Annotated[A | B, strawberry.union("MyUnion")]

    @strawberry.type
    class Query:
        field1: MyUnion
        field2: MyUnion

    schema = strawberry.Schema(query=Query)

    assert schema.as_str() == dedent(
        """\
        type A {
          a: Int!
        }

        type B {
          b: Int!
        }

        union MyUnion = A | B

        type Query {
          field1: MyUnion!
          field2: MyUnion!
        }"""
    )


def test_union_explicit_type_resolution():
    @dataclass
    class ADataclass:
        a: int

    @strawberry.type
    class A:
        a: int

        @classmethod
        def is_type_of(cls, obj, _info) -> bool:
            return isinstance(obj, ADataclass)

    @strawberry.type
    class B:
        b: int

    MyUnion = Annotated[A | B, strawberry.union("MyUnion")]

    @strawberry.type
    class Query:
        @strawberry.field
        def my_field(self) -> MyUnion:
            return ADataclass(a=1)  # type: ignore

    schema = strawberry.Schema(query=Query)

    query = "{ myField { __typename, ... on A { a }, ... on B { b } } }"
    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"myField": {"__typename": "A", "a": 1}}


def test_union_optional_with_or_operator():
    """Verify that the `|` operator is supported when annotating unions as
    optional in schemas.
    """

    @strawberry.type
    class Cat:
        name: str

    @strawberry.type
    class Dog:
        name: str

    animal_union = Annotated[Cat | Dog, strawberry.union("Animal")]

    @strawberry.type
    class Query:
        @strawberry.field
        def animal(self) -> animal_union | None:
            return None

    schema = strawberry.Schema(query=Query)
    query = """{
        animal {
            __typename
        }
    }"""

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["animal"] is None


def test_union_with_input_types():
    """Verify that union of input types raises an error."""

    @strawberry.type
    class User:
        name: str
        age: int

    @strawberry.input
    class A:
        a: str

    @strawberry.input
    class B:
        b: str

    @strawberry.input
    class Input:
        name: str
        something: A | B

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self, data: Input) -> User:
            return User(name=data.name, age=100)

    with pytest.raises(
        TypeError, match="Union for A is not supported because it is an Input type"
    ):
        strawberry.Schema(query=Query)


def test_union_with_similar_nested_generic_types():
    """Previously this failed due to an edge case where Strawberry would choose AContainer
    as the resolved type for container_b due to the inability to exactly match the
    nested generic `Container.items`.
    """
    T = TypeVar("T")

    @strawberry.type
    class Container(Generic[T]):
        items: list[T]

    @strawberry.type
    class A:
        a: str

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class Query:
        @strawberry.field
        def container_a(self) -> Container[A] | A:
            return Container(items=[A(a="hello")])

        @strawberry.field
        def container_b(self) -> Container[B] | B:
            return Container(items=[B(b=3)])

    schema = strawberry.Schema(query=Query)

    query = """
     {
        containerA {
            __typename
            ... on AContainer {
                items {
                    a
                }
            }
            ... on A {
                a
            }
        }
    }
    """

    result = schema.execute_sync(query)

    assert result.data["containerA"]["items"][0]["a"] == "hello"

    query = """
     {
        containerB {
            __typename
            ... on BContainer {
                items {
                    b
                }
            }
            ... on B {
                b
            }
        }
    }
    """

    result = schema.execute_sync(query)

    assert result.data["containerB"]["items"][0]["b"] == 3


def test_union_with_generic_uses_is_type_of_for_domain_objects():
    T = TypeVar("T")

    @dataclass
    class SQLProduct:
        name: str
        price: float

    @strawberry.type
    class BaseGQLType:
        __domain_type__: ClassVar[type | None] = None

        @classmethod
        def is_type_of(cls, obj: Any, _info) -> bool:
            if (cls == obj.__class__) or (
                cls.__domain_type__ and isinstance(obj, cls.__domain_type__)
            ):
                return True
            return (
                super().is_type_of(obj, _info)  # type: ignore[misc]
                if hasattr(super(), "is_type_of")
                else False
            )

    @strawberry.type
    class Product(BaseGQLType):
        __domain_type__: strawberry.Private[type | None] = SQLProduct

        name: str
        price: float

    @strawberry.type
    class Error(BaseGQLType):
        message: str

    @strawberry.type
    class Page(BaseGQLType, Generic[T]):
        items: list[T]

    def get_products() -> Page[Product] | Error:
        return Page[Product](items=[SQLProduct(name="a", price=2.2)])

    @strawberry.type
    class Query:
        products: Page[Product] | Error = strawberry.field(resolver=get_products)

    schema = strawberry.Schema(Query)
    query = """
    {
        products {
            __typename
            ... on ProductPage {
                items {
                    name
                    price
                }
            }
            ... on Error {
                message
            }
        }
    }
    """

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["products"] == {
        "__typename": "ProductPage",
        "items": [{"name": "a", "price": 2.2}],
    }


def test_lazy_union():
    """Previously this failed to evaluate generic parameters on lazy types"""
    TypeA = Annotated["TypeA", lazy("tests.schema.test_lazy_types.type_a")]
    TypeB = Annotated["TypeB", lazy("tests.schema.test_lazy_types.type_b")]

    @strawberry.type
    class Query:
        @strawberry.field
        def a(self) -> TypeA | TypeB:
            from tests.schema.test_lazy_types.type_a import TypeA

            return TypeA(list_of_b=[])

        @strawberry.field
        def b(self) -> TypeA | TypeB:
            from tests.schema.test_lazy_types.type_b import TypeB

            return TypeB()

    schema = strawberry.Schema(query=Query)

    query = """
     {
        a {
            __typename
        }
        b {
            __typename
        }
    }
    """

    result = schema.execute_sync(query)

    assert result.data["a"]["__typename"] == "TypeA"
    assert result.data["b"]["__typename"] == "TypeB"


def test_lazy_union_with_generic():
    UnionValue = Annotated["UnionValue", lazy("tests.schema.test_lazy.type_e")]

    @strawberry.type
    class Query:
        @strawberry.field
        def a(self) -> UnionValue:
            from tests.schema.test_lazy.type_e import MyEnum, ValueContainer

            return ValueContainer(value=MyEnum.ONE)

    schema = strawberry.Schema(query=Query)

    query = """
     {
        a {
            __typename
        }
    }
    """

    result = schema.execute_sync(query)

    assert result.data["a"]["__typename"] == "MyEnumValueContainer"


@pytest.mark.raises_strawberry_exception(
    InvalidUnionTypeError, match="Type `int` cannot be used in a GraphQL Union"
)
def test_error_with_invalid_annotated_type():
    @strawberry.type
    class Something:
        h: str

    AnnotatedInt = Annotated[int, "something_else"]

    @strawberry.type
    class Query:
        union: Something | AnnotatedInt

    strawberry.Schema(query=Query)


@pytest.mark.raises_strawberry_exception(
    InvalidUnionTypeError, match="Type `int` cannot be used in a GraphQL Union"
)
def test_raises_on_union_with_int():
    global ICanBeInUnion

    @strawberry.type
    class ICanBeInUnion:
        foo: str

    @strawberry.type
    class Query:
        union: ICanBeInUnion | int

    strawberry.Schema(query=Query)

    del ICanBeInUnion


@pytest.mark.raises_strawberry_exception(
    InvalidUnionTypeError,
    match=r"Type `list\[...\]` cannot be used in a GraphQL Union",
)
def test_raises_on_union_with_list_str():
    global ICanBeInUnion

    @strawberry.type
    class ICanBeInUnion:
        foo: str

    @strawberry.type
    class Query:
        union: ICanBeInUnion | list[str]

    strawberry.Schema(query=Query)

    del ICanBeInUnion


@pytest.mark.raises_strawberry_exception(
    InvalidUnionTypeError,
    match=r"Type `list\[...\]` cannot be used in a GraphQL Union",
)
def test_raises_on_union_with_list_str_38():
    global ICanBeInUnion

    @strawberry.type
    class ICanBeInUnion:
        foo: str

    @strawberry.type
    class Query:
        union: ICanBeInUnion | list[str]

    strawberry.Schema(query=Query)

    del ICanBeInUnion


@pytest.mark.raises_strawberry_exception(
    InvalidUnionTypeError, match="Type `Always42` cannot be used in a GraphQL Union"
)
def test_raises_on_union_of_custom_scalar():
    @strawberry.type
    class ICanBeInUnion:
        foo: str

    @strawberry.scalar(serialize=lambda x: 42, parse_value=lambda x: Always42())
    class Always42:
        pass

    @strawberry.type
    class Query:
        union: Annotated[
            Always42 | ICanBeInUnion, strawberry.union(name="ExampleUnion")
        ]

    strawberry.Schema(query=Query)


def test_union_of_unions():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Error:
        name: str

    @strawberry.type
    class SpecificError:
        name: str

    @strawberry.type
    class EvenMoreSpecificError:
        name: str

    ErrorUnion = Union[SpecificError, EvenMoreSpecificError]

    @strawberry.type
    class Query:
        user: User | Error
        error: User | ErrorUnion

    schema = strawberry.Schema(query=Query)

    expected_schema = textwrap.dedent(
        """
        type Error {
          name: String!
        }

        type EvenMoreSpecificError {
          name: String!
        }

        type Query {
          user: UserError!
          error: UserSpecificErrorEvenMoreSpecificError!
        }

        type SpecificError {
          name: String!
        }

        type User {
          name: String!
        }

        union UserError = User | Error

        union UserSpecificErrorEvenMoreSpecificError = User | SpecificError | EvenMoreSpecificError
        """
    ).strip()

    assert str(schema) == expected_schema


def test_single_union():
    @strawberry.type
    class A:
        a: int = 5

    @strawberry.type
    class Query:
        something: Annotated[A, strawberry.union(name="Something")] = strawberry.field(
            default_factory=A
        )

    schema = strawberry.Schema(query=Query)
    query = """{
        something {
            __typename,

            ... on A {
                a
            }
        }
    }"""

    assert (
        str(schema)
        == textwrap.dedent(
            """
        type A {
          a: Int!
        }

        type Query {
          something: Something!
        }

        union Something = A
        """
        ).strip()
    )

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["something"] == {"__typename": "A", "a": 5}


def test_generic_union_with_annotated():
    @strawberry.type
    class SomeType:
        id: strawberry.ID
        name: str

    @strawberry.type
    class NotFoundError:
        id: strawberry.ID
        message: str

    T = TypeVar("T")

    @strawberry.type
    class ObjectQueries(Generic[T]):
        @strawberry.field
        def by_id(
            self, id: strawberry.ID
        ) -> Annotated[T | NotFoundError, strawberry.union("ByIdResult")]: ...

    @strawberry.type
    class Query:
        @strawberry.field
        def some_type_queries(self, id: strawberry.ID) -> ObjectQueries[SomeType]:
            raise NotImplementedError

    schema = strawberry.Schema(Query)

    assert (
        str(schema)
        == textwrap.dedent(
            """
            type NotFoundError {
              id: ID!
              message: String!
            }

            type Query {
              someTypeQueries(id: ID!): SomeTypeObjectQueries!
            }

            type SomeType {
              id: ID!
              name: String!
            }

            union SomeTypeByIdResult = SomeType | NotFoundError

            type SomeTypeObjectQueries {
              byId(id: ID!): SomeTypeByIdResult!
            }
            """
        ).strip()
    )


def test_generic_union_with_annotated_inside():
    @strawberry.type
    class SomeType:
        id: strawberry.ID
        name: str

    @strawberry.type
    class NotFoundError:
        id: strawberry.ID
        message: str

    T = TypeVar("T")

    @strawberry.type
    class ObjectQueries(Generic[T]):
        @strawberry.field
        def by_id(
            self, id: strawberry.ID
        ) -> T | Annotated[NotFoundError, strawberry.union("ByIdResult")]: ...

    @strawberry.type
    class Query:
        @strawberry.field
        def some_type_queries(self, id: strawberry.ID) -> ObjectQueries[SomeType]: ...

    schema = strawberry.Schema(Query)

    assert (
        str(schema)
        == textwrap.dedent(
            """
            type NotFoundError {
              id: ID!
              message: String!
            }

            type Query {
              someTypeQueries(id: ID!): SomeTypeObjectQueries!
            }

            type SomeType {
              id: ID!
              name: String!
            }

            union SomeTypeByIdResult = SomeType | NotFoundError

            type SomeTypeObjectQueries {
              byId(id: ID!): SomeTypeByIdResult!
            }
            """
        ).strip()
    )


def test_annoted_union_with_two_generics():
    @strawberry.type
    class SomeType:
        a: str

    @strawberry.type
    class OtherType:
        b: str

    @strawberry.type
    class NotFoundError:
        message: str

    T = TypeVar("T")
    U = TypeVar("U")

    @strawberry.type
    class UnionObjectQueries(Generic[T, U]):
        @strawberry.field
        def by_id(
            self, id: strawberry.ID
        ) -> T | Annotated[U | NotFoundError, strawberry.union("ByIdResult")]: ...

    @strawberry.type
    class Query:
        @strawberry.field
        def some_type_queries(
            self, id: strawberry.ID
        ) -> UnionObjectQueries[SomeType, OtherType]: ...

    schema = strawberry.Schema(Query)

    assert (
        str(schema)
        == textwrap.dedent(
            """
            type NotFoundError {
              message: String!
            }

            type OtherType {
              b: String!
            }

            type Query {
              someTypeQueries(id: ID!): SomeTypeOtherTypeUnionObjectQueries!
            }

            type SomeType {
              a: String!
            }

            union SomeTypeOtherTypeByIdResult = SomeType | OtherType | NotFoundError

            type SomeTypeOtherTypeUnionObjectQueries {
              byId(id: ID!): SomeTypeOtherTypeByIdResult!
            }
            """
        ).strip()
    )


def test_union_merging_without_annotated():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class C:
        c: int

    a = Union[A, B]

    b = Union[B, C]

    c = Union[a, b]

    @strawberry.type
    class Query:
        union_field: c

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == textwrap.dedent(
            """
        type A {
          a: Int!
        }

        union ABC = A | B | C

        type B {
          b: Int!
        }

        type C {
          c: Int!
        }

        type Query {
          unionField: ABC!
        }
    """
        ).strip()
    )


def test_union_merging_with_annotated():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class C:
        c: int

    a = Annotated[A | B, strawberry.union("AorB")]

    b = Annotated[B | C, strawberry.union("BorC")]

    c = Union[a, b]

    @strawberry.type
    class Query:
        union_field: c

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == textwrap.dedent(
            """
        type A {
          a: Int!
        }

        union AorBBorC = A | B | C

        type B {
          b: Int!
        }

        type C {
          c: Int!
        }

        type Query {
          unionField: AorBBorC!
        }
    """
        ).strip()
    )


def test_union_merging_with_annotated_annotated_merge():
    @strawberry.type
    class A:
        a: int

    @strawberry.type
    class B:
        b: int

    @strawberry.type
    class C:
        c: int

    a = Annotated[A | B, strawberry.union("AorB")]

    b = Annotated[B | C, strawberry.union("BorC")]

    c = Annotated[a | b, strawberry.union("ABC")]

    @strawberry.type
    class Query:
        union_field: c

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == textwrap.dedent(
            """
        type A {
          a: Int!
        }

        union ABC = A | B | C

        type B {
          b: Int!
        }

        type C {
          c: Int!
        }

        type Query {
          unionField: ABC!
        }
    """
        ).strip()
    )


def test_union_used_inside_generic():
    T = TypeVar("T")

    @strawberry.type
    class User:
        name: str
        age: int

    @strawberry.type
    class ProUser:
        name: str
        age: float

    @strawberry.type
    class GenType(Generic[T]):
        data: T

    GeneralUser = Annotated[User | ProUser, strawberry.union("GeneralUser")]

    @strawberry.type
    class Response(GenType[GeneralUser]): ...

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> Response: ...

    schema = strawberry.Schema(query=Query)

    assert (
        str(schema)
        == textwrap.dedent(
            """
        union GeneralUser = User | ProUser

        type ProUser {
          name: String!
          age: Float!
        }

        type Query {
          user: Response!
        }

        type Response {
          data: GeneralUser!
        }

        type User {
          name: String!
          age: Int!
        }
    """
        ).strip()
    )


def test_annotated_union_inside_generic_subclass():
    T = TypeVar("T")

    @strawberry.type
    class Dog:
        name: str

    @strawberry.type
    class Cat:
        name: str

    DogCat = Annotated[Union[Dog, Cat], strawberry.union("DogCat")]

    @strawberry.type
    class Listing(Generic[T]):
        items: list[T]

    @strawberry.type
    class Items(Listing[DogCat]):
        pass

    @strawberry.type
    class Query:
        @strawberry.field
        def items(self) -> Items:
            return Items(items=[Dog(name="Fido"), Cat(name="Whiskers")])

    schema = strawberry.Schema(query=Query)

    result = schema.execute_sync("""
        {
            items {
                items {
                    __typename
                    ... on Dog { name }
                    ... on Cat { name }
                }
            }
        }
    """)

    assert not result.errors
    assert result.data == {
        "items": {
            "items": [
                {"__typename": "Dog", "name": "Fido"},
                {"__typename": "Cat", "name": "Whiskers"},
            ]
        }
    }



class OperationNameResultLazy:
    # alias for something
    lazy: bool

class OperationNameResult:
    id: str
    # alias for id
    second_id: str
    # alias for float
    a_float: float
    lazy: OperationNameResultLazy



from contextlib import nullcontext

import pytest

from strawberry.utils.await_maybe import await_maybe


@pytest.mark.parametrize(
    ("assert_no_errors", "expectation"),
    [(True, pytest.raises(AssertionError)), (False, nullcontext())],
)
async def test_query_with_assert_no_errors_option(
    graphql_client, assert_no_errors, expectation
):
    query = "{ ThisIsNotAValidQuery }"

    with expectation:
        await await_maybe(
            graphql_client.query(query, assert_no_errors=assert_no_errors)
        )



from typing import Optional, Union

class AnimalProjection:
    # typename: Animal
    age: int

class OperationNameResultUnionPerson:
    # typename: Person
    name: str

OperationNameResultUnion = Union[AnimalProjection, OperationNameResultUnionPerson]

class OperationNameResultOptionalUnionPerson:
    # typename: Person
    name: str

OperationNameResultOptionalUnion = Union[AnimalProjection, OperationNameResultOptionalUnionPerson]

class OperationNameResult:
    union: OperationNameResultUnion
    optional_union: Optional[OperationNameResultOptionalUnion]



import textwrap

import strawberry


def test_prints_one_of_directive():
    @strawberry.federation.input(one_of=True, tags=["myTag", "anotherTag"])
    class Input:
        a: str | None = strawberry.UNSET
        b: int | None = strawberry.UNSET

    @strawberry.federation.type
    class Query:
        hello: str

    schema = strawberry.federation.Schema(query=Query, types=[Input])

    expected = """
        directive @oneOf on INPUT_OBJECT

        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@tag"]) {
          query: Query
        }

        input Input @tag(name: "myTag") @tag(name: "anotherTag") @oneOf {
          a: String
          b: Int
        }

        type Query {
          _service: _Service!
          hello: String!
        }

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()



from typing import Optional, Union

class OperationNameResultUnionAnimal:
    # typename: Animal
    age: int

class OperationNameResultUnionPerson:
    # typename: Person
    name: str

OperationNameResultUnion = Union[OperationNameResultUnionAnimal, OperationNameResultUnionPerson]

class OperationNameResultOptionalUnionAnimal:
    # typename: Animal
    age: int

class OperationNameResultOptionalUnionPerson:
    # typename: Person
    name: str

OperationNameResultOptionalUnion = Union[OperationNameResultOptionalUnionAnimal, OperationNameResultOptionalUnionPerson]

class OperationNameResult:
    union: OperationNameResultUnion
    optional_union: Optional[OperationNameResultOptionalUnion]



# type: ignore
import enum
import textwrap
from typing import Annotated, Generic, TypeAlias, TypeVar

import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.types.base import get_object_definition
from strawberry.types.field import StrawberryField
from strawberry.types.fields.resolver import StrawberryResolver
from strawberry.types.lazy_type import LazyType
from strawberry.types.union import StrawberryUnion, union

T = TypeVar("T")


# This type is in the same file but should adequately test the logic.
@strawberry.type
class LaziestType:
    something: bool


@strawberry.type
class LazyGenericType(Generic[T]):
    something: T


LazyTypeAlias: TypeAlias = LazyGenericType[int]


@strawberry.enum
class LazyEnum(enum.Enum):
    BREAD = "BREAD"


def test_lazy_type():
    LazierType = LazyType("LaziestType", "tests.types.test_lazy_types")

    annotation = StrawberryAnnotation(LazierType)
    resolved = annotation.resolve()

    # Currently StrawberryAnnotation(LazyType).resolve() returns the unresolved
    # LazyType. We may want to find a way to directly return the referenced object
    # without a second resolving step.
    assert isinstance(resolved, LazyType)
    assert resolved is LazierType
    assert resolved.resolve_type() is LaziestType


def test_lazy_type_alias():
    LazierType = LazyType("LazyTypeAlias", "tests.types.test_lazy_types")

    annotation = StrawberryAnnotation(LazierType)
    resolved = annotation.resolve()

    # Currently StrawberryAnnotation(LazyType).resolve() returns the unresolved
    # LazyType. We may want to find a way to directly return the referenced object
    # without a second resolving step.
    assert isinstance(resolved, LazyType)
    resolved_type = resolved.resolve_type()
    assert resolved_type.__origin__ is LazyGenericType
    assert resolved_type.__args__ == (int,)


def test_lazy_type_function():
    LethargicType = Annotated[
        "LaziestType", strawberry.lazy("tests.types.test_lazy_types")
    ]

    annotation = StrawberryAnnotation(LethargicType)
    resolved = annotation.resolve()

    assert isinstance(resolved, LazyType)
    assert resolved.resolve_type() is LaziestType


def test_lazy_type_enum():
    LazierType = LazyType("LazyEnum", "tests.types.test_lazy_types")

    annotation = StrawberryAnnotation(LazierType)
    resolved = annotation.resolve()

    # Currently StrawberryAnnotation(LazyType).resolve() returns the unresolved
    # LazyType. We may want to find a way to directly return the referenced object
    # without a second resolving step.
    assert isinstance(resolved, LazyType)
    assert resolved is LazierType
    assert resolved.resolve_type() is LazyEnum


def test_lazy_type_argument():
    LazierType = LazyType("LaziestType", "tests.types.test_lazy_types")

    @strawberry.mutation
    def slack_off(emotion: LazierType) -> bool:
        _ = emotion
        return True

    argument = slack_off.arguments[0]
    assert isinstance(argument.type, LazyType)
    assert argument.type is LazierType
    assert argument.type.resolve_type() is LaziestType


def test_lazy_type_field():
    LazierType = LazyType("LaziestType", "tests.types.test_lazy_types")

    annotation = StrawberryAnnotation(LazierType)
    field = StrawberryField(type_annotation=annotation)

    assert isinstance(field.type, LazyType)
    assert field.type is LazierType
    assert field.type.resolve_type() is LaziestType


def test_lazy_type_generic():
    T = TypeVar("T")

    @strawberry.type
    class GenericType(Generic[T]):
        item: T

    LazierType = LazyType("LaziestType", "tests.types.test_lazy_types")
    ResolvedType = GenericType[LazierType]

    annotation = StrawberryAnnotation(ResolvedType)
    resolved = annotation.resolve()

    definition = get_object_definition(resolved)
    assert definition
    items_field: StrawberryField = definition.fields[0]
    assert items_field.type is LazierType
    assert items_field.type.resolve_type() is LaziestType


def test_lazy_type_object():
    LazierType = LazyType("LaziestType", "tests.types.test_lazy_types")

    @strawberry.type
    class WaterParkFeature:
        river: LazierType

    field: StrawberryField = WaterParkFeature.__strawberry_definition__.fields[0]

    assert isinstance(field.type, LazyType)
    assert field.type is LazierType
    assert field.type.resolve_type() is LaziestType


def test_lazy_type_resolver():
    LazierType = LazyType("LaziestType", "tests.types.test_lazy_types")

    def slaking_pokemon() -> LazierType:
        raise NotImplementedError

    resolver = StrawberryResolver(slaking_pokemon)
    assert isinstance(resolver.type, LazyType)
    assert resolver.type is LazierType
    assert resolver.type.resolve_type() is LaziestType


def test_lazy_type_in_union():
    ActiveType = LazyType("LaziestType", "tests.types.test_lazy_types")
    ActiveEnum = LazyType("LazyEnum", "tests.types.test_lazy_types")

    something = Annotated[ActiveType | ActiveEnum, union(name="CoolUnion")]
    annotation = StrawberryAnnotation(something)

    resolved = annotation.resolve()
    assert isinstance(resolved, StrawberryUnion)

    [type1, type2] = resolved.types
    assert type1 is ActiveType
    assert type2 is ActiveEnum
    assert type1.resolve_type() is LaziestType
    assert type2.resolve_type() is LazyEnum


def test_lazy_function_in_union():
    ActiveType = Annotated[
        "LaziestType", strawberry.lazy("tests.types.test_lazy_types")
    ]
    ActiveEnum = Annotated["LazyEnum", strawberry.lazy("tests.types.test_lazy_types")]

    something = Annotated[ActiveType | ActiveEnum, union(name="CoolUnion")]
    annotation = StrawberryAnnotation(something)

    resolved = annotation.resolve()
    assert isinstance(resolved, StrawberryUnion)

    [type1, type2] = resolved.types
    assert type1.resolve_type() is LaziestType
    assert type2.resolve_type() is LazyEnum


def test_optional_lazy_type_using_or_operator():
    from tests.schema.test_lazy.type_a import TypeA

    @strawberry.type
    class SomeType:
        foo: Annotated[TypeA, strawberry.lazy("tests.schema.test_lazy.type_a")] | None

    @strawberry.type
    class AnotherType:
        foo: TypeA | None = None

    @strawberry.type
    class Query:
        some_type: SomeType
        another_type: AnotherType

    schema = strawberry.Schema(query=Query)
    expected = """\
    type AnotherType {
      foo: TypeA
    }

    type Query {
      someType: SomeType!
      anotherType: AnotherType!
    }

    type SomeType {
      foo: TypeA
    }

    type TypeA {
      listOfB: [TypeB!]
      typeB: TypeB!
    }

    type TypeB {
      typeA: TypeA!
      typeAList: [TypeA!]!
      typeCList: [TypeC!]!
    }

    type TypeC {
      name: String!
    }
    """
    assert str(schema).strip() == textwrap.dedent(expected).strip()



import textwrap
from dataclasses import InitVar, dataclass
from enum import Enum

import pytest

import strawberry
from strawberry import ID
from strawberry.scalars import Base64
from strawberry.schema_directive import Location
from strawberry.types.base import StrawberryList


def test_raises_exception_with_unsupported_types():
    class SomeType: ...

    @strawberry.type
    class Query:
        example: SomeType

    with pytest.raises(
        TypeError, match=r"Query fields cannot be resolved. Unexpected type '.*'"
    ):
        strawberry.Schema(query=Query)


def test_basic_schema():
    @strawberry.type
    class Query:
        example: str = "Example"

    schema = strawberry.Schema(query=Query)

    query = "{ example }"

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["example"] == "Example"


def test_basic_schema_optional():
    @strawberry.type
    class Query:
        example: str | None = None

    schema = strawberry.Schema(query=Query)

    query = "{ example }"

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["example"] is None


def test_basic_schema_types():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Query:
        user: User | None = None

    schema = strawberry.Schema(query=Query)

    query = "{ user { name } }"

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["user"] is None


def test_does_camel_case_conversion():
    @strawberry.type
    class Query:
        @strawberry.field
        def hello_world(self, query_param: str) -> str:
            return query_param

    schema = strawberry.Schema(query=Query)

    query = """{
        helloWorld(queryParam: "hi")
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["helloWorld"] == "hi"


def test_can_rename_fields():
    @strawberry.type
    class Hello:
        value: str | None = strawberry.field(name="name")

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> Hello:
            return Hello(value="hi")

        @strawberry.field(name="example1")
        def example(self, query_param: str) -> str:
            return query_param

    schema = strawberry.Schema(query=Query)

    query = """{
        hello { name }
        example1(queryParam: "hi")
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["hello"]["name"] == "hi"
    assert result.data["example1"] == "hi"


def test_type_description():
    @strawberry.type(description="Decorator argument description")
    class TypeA:
        a: str

    @strawberry.type
    class Query:
        a: TypeA

    schema = strawberry.Schema(query=Query)

    query = """{
        __type(name: "TypeA") {
            name
            description
        }
    }"""

    result = schema.execute_sync(query)
    assert not result.errors

    assert result.data["__type"] == {
        "name": "TypeA",
        "description": "Decorator argument description",
    }


def test_field_description():
    @strawberry.type
    class Query:
        a: str = strawberry.field(description="Example")

        @strawberry.field
        def b(self, id: int) -> str:
            return "I'm a resolver"

        @strawberry.field(description="Example C")
        def c(self, id: int) -> str:
            return "I'm a resolver"

    schema = strawberry.Schema(query=Query)

    query = """{
        __type(name: "Query") {
            fields {
                name
                description
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors

    assert result.data["__type"]["fields"] == [
        {"name": "a", "description": "Example"},
        {"name": "b", "description": None},
        {"name": "c", "description": "Example C"},
    ]


def test_field_deprecated_reason():
    @strawberry.type
    class Query:
        a: str = strawberry.field(deprecation_reason="Deprecated A")

        @strawberry.field
        def b(self, id: int) -> str:
            return "I'm a resolver"

        @strawberry.field(deprecation_reason="Deprecated B")
        def c(self, id: int) -> str:
            return "I'm a resolver"

    schema = strawberry.Schema(query=Query)

    query = """{
        __type(name: "Query") {
            fields(includeDeprecated: true) {
                name
                deprecationReason
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["__type"]["fields"] == [
        {"name": "a", "deprecationReason": "Deprecated A"},
        {"name": "b", "deprecationReason": None},
        {"name": "c", "deprecationReason": "Deprecated B"},
    ]


def test_field_deprecated_reason_subscription():
    @strawberry.type
    class Query:
        a: str

    @strawberry.type
    class Subscription:
        @strawberry.subscription(deprecation_reason="Deprecated A")
        def a(self) -> str:
            return "A"

    schema = strawberry.Schema(query=Query, subscription=Subscription)

    query = """{
        __type(name: "Subscription") {
            fields(includeDeprecated: true) {
                name
                deprecationReason
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["__type"]["fields"] == [
        {"name": "a", "deprecationReason": "Deprecated A"},
    ]


def test_enum_description():
    @strawberry.enum(description="We love ice-creams")
    class IceCreamFlavour(Enum):
        VANILLA = "vanilla"
        STRAWBERRY = "strawberry"
        CHOCOLATE = "chocolate"

    @strawberry.enum
    class PizzaType(Enum):
        MARGHERITA = "margherita"

    @strawberry.type
    class Query:
        favorite_ice_cream: IceCreamFlavour = IceCreamFlavour.STRAWBERRY
        pizza: PizzaType = PizzaType.MARGHERITA

    schema = strawberry.Schema(query=Query)

    query = """{
        iceCreamFlavour: __type(name: "IceCreamFlavour") {
            description
            enumValues {
                name
                description
            }
        }
        pizzas: __type(name: "PizzaType") {
            description
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors

    assert result.data["iceCreamFlavour"]["description"] == "We love ice-creams"
    assert result.data["iceCreamFlavour"]["enumValues"] == [
        {"name": "VANILLA", "description": None},
        {"name": "STRAWBERRY", "description": None},
        {"name": "CHOCOLATE", "description": None},
    ]

    assert result.data["pizzas"]["description"] is None


def test_enum_value_description():
    @strawberry.enum
    class IceCreamFlavour(Enum):
        VANILLA = "vainilla"
        STRAWBERRY = strawberry.enum_value("strawberry", description="Our favourite.")
        CHOCOLATE = "chocolate"

    @strawberry.type
    class Query:
        favorite_ice_cream: IceCreamFlavour = IceCreamFlavour.STRAWBERRY

    schema = strawberry.Schema(query=Query)

    query = """{
        iceCreamFlavour: __type(name: "IceCreamFlavour") {
            enumValues {
                name
                description
            }
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["iceCreamFlavour"]["enumValues"] == [
        {"name": "VANILLA", "description": None},
        {"name": "STRAWBERRY", "description": "Our favourite."},
        {"name": "CHOCOLATE", "description": None},
    ]


def test_parent_class_fields_are_inherited():
    @strawberry.type
    class Parent:
        cheese: str = "swiss"

        @strawberry.field
        def friend(self) -> str:
            return "food"

    @strawberry.type
    class Schema(Parent):
        cake: str = "made_in_switzerland"

        @strawberry.field
        def hello_this_is(self) -> str:
            return "patrick"

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> Schema:
            return Schema()

    schema = strawberry.Schema(query=Query)

    query = "{ example { cheese, cake, friend, helloThisIs } }"

    result = schema.execute_sync(query)

    assert not result.errors

    assert result.data["example"]["cheese"] == "swiss"
    assert result.data["example"]["cake"] == "made_in_switzerland"
    assert result.data["example"]["friend"] == "food"
    assert result.data["example"]["helloThisIs"] == "patrick"


def test_can_return_compatible_type():
    """Test that we can return a different type that has the same fields,
    for example when returning a Django Model.
    """

    @dataclass
    class Example:
        name: str

    @strawberry.type
    class Cheese:
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def assortment(self) -> Cheese:
            return Example(name="Asiago")  # type: ignore

    schema = strawberry.Schema(query=Query)

    query = """{
        assortment {
            name
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["assortment"]["name"] == "Asiago"


def test_init_var():
    @strawberry.type
    class Category:
        name: str
        id: InitVar[str]

    @strawberry.type
    class Query:
        @strawberry.field
        def category(self) -> Category:
            return Category(name="example", id="123")

    schema = strawberry.Schema(query=Query)

    query = "{ category { name } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["category"]["name"] == "example"


def test_nested_types():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User(name="Patrick")

    schema = strawberry.Schema(query=Query)

    query = "{ user { name } }"

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data["user"]["name"] == "Patrick"


def test_multiple_fields_with_same_type():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Query:
        me: User | None = None
        you: User | None = None

    schema = strawberry.Schema(query=Query)

    query = "{ me { name } you { name } }"

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data["me"] is None
    assert result.data["you"] is None


def test_str_magic_method_prints_schema_sdl():
    @strawberry.type
    class Query:
        example_bool: bool
        example_str: str = "Example"
        example_int: int = 1

    schema = strawberry.Schema(query=Query)
    expected = """
    type Query {
      exampleBool: Boolean!
      exampleStr: String!
      exampleInt: Int!
    }
    """
    assert str(schema) == textwrap.dedent(expected).strip()
    assert "<strawberry.schema.schema.Schema object" in repr(schema), (
        "Repr should not be affected"
    )


def test_field_with_default():
    @strawberry.type
    class Query:
        a: str = strawberry.field(default="Example")

    schema = strawberry.Schema(query=Query)

    query = "{ a }"

    result = schema.execute_sync(query, root_value=Query())

    assert not result.errors
    assert result.data == {"a": "Example"}


def test_with_types():
    # Ensures Schema(types=[...]) works with all data types
    @strawberry.type
    class Type:
        foo: int

    @strawberry.interface
    class Interface:
        foo: int

    @strawberry.input
    class Input:
        foo: int

    @strawberry.type
    class Query:
        foo: int

    @strawberry.schema_directive(locations=[Location.SCALAR], name="specifiedBy")
    class SpecifiedBy:
        name: str

    schema = strawberry.Schema(
        query=Query, types=[Type, Interface, Input, Base64, ID, str, int, SpecifiedBy]
    )
    expected = '''
        directive @specifiedBy(name: String!) on SCALAR

        """
        Represents binary data as Base64-encoded strings, using the standard alphabet.
        """
        scalar Base64 @specifiedBy(url: "https://datatracker.ietf.org/doc/html/rfc4648.html#section-4")

        input Input {
          foo: Int!
        }

        interface Interface {
          foo: Int!
        }

        type Query {
          foo: Int!
        }

        type Type {
          foo: Int!
        }
    '''

    assert str(schema) == textwrap.dedent(expected).strip()


def test_with_types_non_named():
    @strawberry.type
    class Query:
        foo: int

    with pytest.raises(TypeError, match=r"\[Int!\] is not a named GraphQL Type"):
        strawberry.Schema(query=Query, types=[StrawberryList(int)])


def test_kw_only():
    @strawberry.type
    class FooBar1:
        foo: int = 1
        bar: int

    @strawberry.type
    class FooBar2:
        foo: int = strawberry.field(default=1)
        bar: int = strawberry.field()

    for FooBar in (FooBar1, FooBar2):
        with pytest.raises(
            TypeError, match="missing 1 required keyword-only argument: 'bar'"
        ):
            FooBar()
        with pytest.raises(
            TypeError, match="missing 1 required keyword-only argument: 'bar'"
        ):
            FooBar(foo=1)
        FooBar(bar=2)
        FooBar(foo=1, bar=2)


def test_execute_sync_with_aiter_object():
    """Sync execution works with objects that define __aiter__.

    graphql-core 3.3's is_async_iterable default treats any object with
    __aiter__ (e.g. Django QuerySets) as async, causing execute_sync to
    return a coroutine instead of an ExecutionResult.
    """

    class FakeQuerySet:
        def __aiter__(self):
            raise NotImplementedError

        def __iter__(self):
            yield from [{"name": "Alice"}, {"name": "Bob"}]

    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Query:
        @strawberry.field
        def users(self) -> list[User]:
            return [User(name=u["name"]) for u in FakeQuerySet()]

    schema = strawberry.Schema(query=Query)
    result = schema.execute_sync("{ users { name } }")

    assert not result.errors
    assert result.data == {"users": [{"name": "Alice"}, {"name": "Bob"}]}



import logging

from graphql.error import GraphQLError

from strawberry.utils.logging import StrawberryLogger


def test_strawberry_logger_error(caplog):
    caplog.set_level(logging.ERROR, logger="strawberry.execution")

    exc = GraphQLError("test exception")
    StrawberryLogger.error(exc)

    assert caplog.record_tuples == [
        ("strawberry.execution", logging.ERROR, "test exception")
    ]



import textwrap

from strawberry.schema_codegen import codegen


def test_support_for_key_directive():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.0", import: ["@key"])

    type User @key(fields: "id") {
        id: ID!
        username: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.federation.type(keys=["id"])
        class User:
            id: strawberry.ID
            username: str
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_support_for_shareable_directive():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.0", import: ["@shareable"])

    type User @shareable {
        id: ID!
        username: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.federation.type(shareable=True)
        class User:
            id: strawberry.ID
            username: str
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_support_for_inaccessible_directive():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.0", import: ["@inaccessible"])

    type User @inaccessible {
        id: ID!
        username: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.federation.type(inaccessible=True)
        class User:
            id: strawberry.ID
            username: str
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_support_for_tags_directive():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.0", import: ["@tags"])

    type User @tag(name: "user") @tag(name: "admin") {
        id: ID!
        username: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.federation.type(tags=["user", "admin"])
        class User:
            id: strawberry.ID
            username: str
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_uses_federation_schema():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.0", import: ["@inaccessible"])

    type Query {
        me: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Query:
            me: str

        schema = strawberry.federation.Schema(query=Query)
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_supports_authenticated_directive():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@authenticated"])

    type User @authenticated {
        name: String! @authenticated
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.federation.type(authenticated=True)
        class User:
            name: str = strawberry.federation.field(authenticated=True)
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_requires_scope():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@requiresScope"])

    type User @requiresScopes(scopes: [["client", "poweruser"], ["admin"], ["productowner"]]){
        name: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.federation.type(requires_scopes=[["client", "poweruser"], ["admin"], ["productowner"]])
        class User:
            name: str
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_policy_directive():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@policy"])

    type User @policy(policies: ["userPolicy", [["client", "poweruser"], ["admin"], ["productowner"]]]){
        name: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.federation.type(policy=["userPolicy", [["client", "poweruser"], ["admin"], ["productowner"]]])
        class User:
            name: str
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_support_for_directives_on_fields():
    schema = """
    extend schema @link(url: "https://specs.apollo.dev/federation/v2.0", import: ["@requires", "@provides"])

    type User {
        a: String! @shareable
        b: String! @inaccessible
        c: String! @override(from: "mySubGraph")
        c1: String! @override(from: "mySubGraph", label: "some.label")
        d: String! @external
        e: String! @requires(fields: "id")
        f: String! @provides(fields: "id")
        g: String! @tag(name: "user")
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry
        from strawberry.federation.schema_directives import Override

        @strawberry.type
        class User:
            a: str = strawberry.federation.field(shareable=True)
            b: str = strawberry.federation.field(inaccessible=True)
            c: str = strawberry.federation.field(override="mySubGraph")
            c1: str = strawberry.federation.field(override=Override(override_from="mySubGraph", label="some.label"))
            d: str = strawberry.federation.field(external=True)
            e: str = strawberry.federation.field(requires=["id"])
            f: str = strawberry.federation.field(provides=["id"])
            g: str = strawberry.federation.field(tags=["user"])
        """
    ).strip()

    assert codegen(schema).strip() == expected



from pathlib import Path

import pytest
from pytest_codspeed.plugin import BenchmarkFixture

from .api import schema, schema_with_directives

ROOT = Path(__file__).parent / "queries"

basic_query = (ROOT / "simple.graphql").read_text()
many_fields_query = (ROOT / "many_fields.graphql").read_text()
many_fields_query_directives = (ROOT / "many_fields_directives.graphql").read_text()
items_query = (ROOT / "items.graphql").read_text()


@pytest.mark.benchmark
def test_execute_basic(benchmark: BenchmarkFixture):
    benchmark(schema.execute_sync, basic_query)


@pytest.mark.benchmark
def test_execute_with_many_fields(benchmark: BenchmarkFixture):
    benchmark(schema.execute_sync, many_fields_query)


@pytest.mark.benchmark
def test_execute_with_many_fields_and_directives(benchmark: BenchmarkFixture):
    benchmark(schema_with_directives.execute_sync, many_fields_query_directives)


@pytest.mark.benchmark
def test_execute_with_10_items(benchmark: BenchmarkFixture):
    benchmark(schema.execute_sync, items_query, variable_values={"count": 10})


@pytest.mark.benchmark
def test_execute_with_100_items(benchmark: BenchmarkFixture):
    benchmark(schema.execute_sync, items_query, variable_values={"count": 100})


@pytest.mark.benchmark
def test_execute_with_1000_items(benchmark: BenchmarkFixture):
    benchmark(schema.execute_sync, items_query, variable_values={"count": 1000})



"""These tests are for Generics that don't expose any generic parts to the schema."""

import sys
import textwrap

import pytest

import strawberry

pytestmark = pytest.mark.skipif(
    sys.version_info < (3, 12), reason="These are tests for Python 3.12+"
)


def test_does_not_create_a_new_type_when_no_generic_field_exposed():
    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID
        node_field: strawberry.Private[T]

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> Edge[int]:
            return Edge(cursor=strawberry.ID("1"), node_field=1)

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename
            cursor
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"example": {"__typename": "Edge", "cursor": "1"}}


def test_does_not_create_a_new_type_when_no_generic_field_exposed_argument():
    @strawberry.type
    class Edge[T]:
        cursor: strawberry.ID

        def something(input: T) -> T:
            return input

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> Edge[int]:
            return Edge(cursor=strawberry.ID("1"))

    schema = strawberry.Schema(query=Query)

    query = """{
        example {
            __typename
            cursor
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data == {"example": {"__typename": "Edge", "cursor": "1"}}


def test_with_interface():
    @strawberry.interface
    class GenericInterface[T]:
        data: strawberry.Private[T]

        @strawberry.field
        def value(self) -> str:
            raise NotImplementedError

    @strawberry.type
    class ImplementationOne(GenericInterface[str]):
        @strawberry.field
        def value(self) -> str:
            return self.data

    @strawberry.type
    class ImplementationTwo(GenericInterface[bool]):
        @strawberry.field
        def value(self) -> str:
            if self.data is True:
                return "true"
            return "false"

    @strawberry.type
    class Query:
        @strawberry.field
        @staticmethod
        async def generic_field() -> GenericInterface:  # type: ignore
            return ImplementationOne(data="foo")

    schema = strawberry.Schema(
        query=Query, types=[ImplementationOne, ImplementationTwo]
    )

    expected_schema = textwrap.dedent(
        """
        interface GenericInterface {
          value: String!
        }

        type ImplementationOne implements GenericInterface {
          value: String!
        }

        type ImplementationTwo implements GenericInterface {
          value: String!
        }

        type Query {
          genericField: GenericInterface!
        }
        """
    ).strip()

    assert str(schema) == expected_schema


def test_with_interface_and_type():
    @strawberry.interface
    class GenericInterface[T]:
        data: strawberry.Private[T]

        @strawberry.field
        def value(self) -> str:
            raise NotImplementedError

    @strawberry.type
    class ImplementationOne(GenericInterface[str]):
        @strawberry.field
        def value(self) -> str:
            return self.data

    @strawberry.type
    class ImplementationTwo(GenericInterface[bool]):
        @strawberry.field
        def value(self) -> str:
            if self.data is True:
                return "true"
            return "false"

    @strawberry.type
    class Query:
        @strawberry.field
        @staticmethod
        async def generic_field() -> GenericInterface[str | bool]:
            return ImplementationOne(data="foo")

    schema = strawberry.Schema(
        query=Query, types=[ImplementationOne, ImplementationTwo]
    )

    expected_schema = textwrap.dedent(
        """
        interface GenericInterface {
          value: String!
        }

        type ImplementationOne implements GenericInterface {
          value: String!
        }

        type ImplementationTwo implements GenericInterface {
          value: String!
        }

        type Query {
          genericField: GenericInterface!
        }
        """
    ).strip()

    assert str(schema) == expected_schema



from __future__ import annotations

from typing import Literal

import pytest

from strawberry.http import GraphQLHTTPResponse
from strawberry.types import ExecutionResult
from tests.views.schema import schema

from .clients.base import HttpClient


def process_result(result: ExecutionResult) -> GraphQLHTTPResponse:
    if result.data:
        return {
            "data": {key.upper(): result for key, result in result.data.items()},
        }

    return {}


@pytest.fixture
def http_client(http_client_class) -> HttpClient:
    return http_client_class(schema, result_override=process_result)


@pytest.mark.parametrize("method", ["get", "post"])
async def test_custom_process_result(
    method: Literal["get", "post"], http_client: HttpClient
):
    response = await http_client.query(
        method=method,
        query="{ hello }",
    )
    assert response.json["data"] == {"HELLO": "Hello world"}



import strawberry


def test_field_name_standard():
    standard_field = strawberry.field()

    assert standard_field.python_name is None
    assert standard_field.graphql_name is None


def test_field_name_standard_on_schema():
    @strawberry.type()
    class Query:
        normal_field: int

    [field] = Query.__strawberry_definition__.fields

    assert field.python_name == "normal_field"
    assert field.graphql_name is None


def test_field_name_override():
    field_name = "override"

    standard_field = strawberry.field(name=field_name)

    assert standard_field.python_name is None  # Set once field is added to a Schema
    assert standard_field.graphql_name == field_name


def test_field_name_override_with_schema():
    field_name = "override_name"

    @strawberry.type()
    class Query:
        override_field: bool = strawberry.field(name=field_name)

    [field] = Query.__strawberry_definition__.fields

    assert field.python_name == "override_field"
    assert field.graphql_name == field_name
